"""
A stemmer is a function from a snippet to a list of strings, where each string represents a stemmed form of
the snippet. A stemming function may produce multiple stemmed forms for a given snippet.

Internally, most stemming functions work on either individual notes or pairwise between notes.
"""

from itertools import islice, chain
from operator import itemgetter
from firms.models import flatten
from music21.interval import Interval
from music21.chord import Chord
from music21.note import Note

def window(seq, window_size=2):
    """
    Returns a sliding window (of width n) over data from the iterable"
       s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
    arguments:
    seq         the sequence to window over
    window_size the size of the windows
    """
    iterator = iter(seq)
    result = tuple(islice(iterator, window_size))
    if len(result) == window_size:
        yield result
    for elem in iterator:
        result = result[1:] + (elem,)
        yield result

def get_number_of_voices(gnote):
    return len(gnote.pitches)

def split_voices(lead, current):
    """
    Given two sequential notes, if the lead note contains more voices than current, then duplicate
    the notes in current to match lead, producing continuous voice lines
        :param lead: Leading note
        :param current: Current node
    """
    num_lead = get_number_of_voices(lead)
    num_current = get_number_of_voices(current)
    if current.isNote:
        return [ Note(current.pitch, quarterLength=current.duration.quarterLength) for i in range(num_lead) ]
    if num_lead == num_current:
        return [ Note(pitch, quarterLength=current.duration.quarterLength) for pitch in current.pitches]
    if current.isChord and len(current.pitches) == 0:
        raise RuntimeError("The system was able to parse the file, but detected an illegal construct: empty chord")
    middle = map(itemgetter(0), [sorted([ (c, abs(Interval(e,c).cents)) for c in current.pitches ], key=itemgetter(1))[0] for e in lead.pitches[1:-1]])
    return [ Note(pitch, quarterLength=current.duration.quarterLength) for pitch in chain(current.pitches[0:1], middle, current.pitches[-1:])]

def split_voice_lines(indexed_notes):
    """
    Split potentially polyphonic line into multiple monophonic lines using voice leading

    indexed_notes   List of tuples of the form (position, GeneralNote)
    
    returns         List of lists of tuples of the form (position, Note)
    """
    max_number_of_voices = max(get_number_of_voices(note) for idx,note in indexed_notes)
    peak = [i for i,(idx,note) in enumerate(indexed_notes) if get_number_of_voices(note) == max_number_of_voices][0]

    head = indexed_notes[:peak+1][::-1]
    climb = []
    if len(head) > 1:
        lead = head[0][1]
        for i,n in head[1:]:
            split_result = split_voices(lead, n)
            climb.append( (i, split_result) )
            lead = Chord(split_result)
    
    tail = indexed_notes[peak:]
    fall = []
    if len(tail) > 1:
        lead = tail[0][1]
        for i,n in tail[1:]:
            split_result = split_voices(lead, n)
            fall.append( (i, split_result) )
            lead = Chord(split_result)

    results = []

    for i in range(max_number_of_voices):
        results.append([])

    source_idx_for_peak,peak_note = indexed_notes[peak]
    for i,result in enumerate(results):
        for j,voices in climb:
            result.append( (j, voices[i]) )
        for j,voices in fall:
            result.append( (j, voices[i]) )
        result.append( (source_idx_for_peak, Note(peak_note.pitches[i], quarterLength=peak_note.duration.quarterLength) ) )

    return results
    
def get_voice_lines(notes):
    """
    Breaks up a stream of potentially polyphonic notes into one or more monophonic (or unphonated) voices
    notes   stream of GeneralNote objects
    """
    # If everything is a rest, just wrap the line in a list and return as is
    if all(note.isRest or note.isNote for note in notes):
        return [notes]
    indexed_notes = list(enumerate(notes))
    indexed_non_rests = [(idx,note) for idx,note in indexed_notes if not note.isRest]
    indexed_rests = [(idx,note) for idx,note in indexed_notes if note.isRest]

    # Do voice leading IGNORING rests
    # List[ List[ (idx, note.Note) ] ]
    raw_voice_lines = split_voice_lines(indexed_non_rests)

    # Mix the rests back in
    voice_lines = []
    for rvl in raw_voice_lines:
        with_rests = chain(rvl, indexed_rests)
        ordered = sorted(with_rests, key=itemgetter(0))
        without_idx = [o[1] for o in ordered]
        voice_lines.append(without_idx)

    return voice_lines

def get_contour(note1, note2):
    if note1.isRest and note2.isRest:
        return 's'
    elif note1.isRest and not note2.isRest:
        return 'u'
    elif not note1.isRest and note2.isRest:
        return 'd'
    else:
        interval = Interval(note1, note2)
        cents = interval.cents
        if cents == 0:
            return 's'
        if cents > 0:
            return 'u'
        return 'd'

def get_interval(note1, note2):
    if note1.isRest or note2.isRest:
        return 'rest'
    return str(Interval(note1, note2).cents)

def stringify_keys(key_list):
    """
    Transform a list of stem keys into a list of string keys for persistence
        :param key_list: List of stem keys
    """
    return [ [str(item) for item in l] for l in key_list]

def join_stem_by_note(note_stems):
    return [ ' '.join(stem) for stem in note_stems ]

def stem_by_pitch(snippet):
    voices = get_voice_lines(snippet.notes)
    return [
        [note.pitch.nameWithOctave if note.isNote else
        "[ %s ]" % ' '.join([pitch.nameWithOctave for pitch in note.pitches])  if note.isChord else
        'rest'
        for note in voice] for voice in voices
    ]

def index_key_by_pitch(snippet):
    return join_stem_by_note(stem_by_pitch(snippet))

def stem_by_simple_pitch(snippet):
    voices = get_voice_lines(snippet.notes)
    return [
        [note.pitch.name if note.isNote else
        "[ %s ]" % ' '.join([pitch.name for pitch in note.pitches]) if note.isChord else
        'rest'
        for note in voice]for voice in voices
    ]

def index_key_by_simple_pitch(snippet):
    return join_stem_by_note(stem_by_simple_pitch(snippet))

def stem_by_interval(snippet):
    voices = get_voice_lines(snippet.notes)
    return [
        [
            get_interval(*pair) for pair in window(voice)
        ] for voice in voices
    ]

def index_key_by_interval(snippet):
    return join_stem_by_note(stem_by_interval(snippet))

def stem_by_contour(snippet):
    voices = get_voice_lines(snippet.notes)
    return [
        [
            get_contour(*pair) for pair in window(voice)
        ] for voice in voices
    ]

def index_key_by_contour(snippet):
    return join_stem_by_note(stem_by_contour(snippet))

def stem_by_rythm(snippet):
    return [
        [ note.duration.quarterLength for note in snippet.notes ]
    ]

def index_key_by_rythm(snippet):
    return join_stem_by_note(stringify_keys(stem_by_rythm(snippet)))

def stem_by_normalized_rythm(snippet):
    rythm_stems = stem_by_rythm(snippet)
    longest = max(flatten(rythm_stems))
    if longest == 0:
        longest = 1.0
    return [ [stem / longest for stem in rythm_stem] for rythm_stem in rythm_stems ]

def index_key_by_normalized_rythm(snippet):
    return join_stem_by_note(stringify_keys(stem_by_normalized_rythm(snippet)))


import unittest
from music21 import converter

from firms.models import Snippet
from firms.stemmers import stem_by_normalized_rythm

QUARTER_RYTHM_LINE = "tinynotation: a b c d e"
VARIABLE_RYTHM_LINE = "tinynotation: a1 b4 c d e"

def build_simple_snippet(notes):
    return Snippet('test', 'test', notes, 0)

class TestStemByNormalizedRythm(unittest.TestCase):
    def setUp(self):
        self.QUARTER_RYTHM_LINE = build_simple_snippet(converter.parse(QUARTER_RYTHM_LINE).flat.notes)
        self.VARIABLE_RYTHM_LINE = build_simple_snippet(converter.parse(VARIABLE_RYTHM_LINE).flat.notes)

    def test_uniform_line(self):
        stemmed = stem_by_normalized_rythm(self.QUARTER_RYTHM_LINE)
        for stem in stemmed:
            for item in stem:
                self.assertEqual(item, 1.0)
    
    def test_VARIABLE_RYTHM_LINE(self):
        stemmed = stem_by_normalized_rythm(self.VARIABLE_RYTHM_LINE)
        expected = [ 1.0, 0.25, 0.25, 0.25, 0.25 ]
        for stem in stemmed:
            self.assertListEqual(stem, expected)

if __name__ == '__main__':
    unittest.main()


from music21 import stream, converter

def wrap_query_as_piece(query):
    piece = stream.Score()
    part = stream.Part()
    part.partName = "query"
    measures = converter.parse("tinynotation: %s" % query)
    for measure in measures:
        part.append(measure)
    piece.append(part)
    return piece


"""
A Sqlite3 based implementation of FIRMS
"""

from itertools import chain
import sqlite3

from music21.repeat import ExpanderException

from firms.models import IRSystem, FirmIndex, get_part_details, get_snippets_for_part

class SqlIRSystem(IRSystem):
    """
    A Sqlite3 based implementation of IRSystem
        :param IRSystem:
    """
    def __init__(self, dbpath, index_methods, graders=None, piece_paths=None, rebuild=True):
        piece_paths = piece_paths or []
        self.dbpath = dbpath
        with sqlite3.connect(self.dbpath) as conn:
            self.ensure_db(conn)
            self.stemmer_ids = self.ensure_stemmers(index_methods, conn)
        super().__init__(index_methods, graders, piece_paths, rebuild)

    def make_empty_index(self, indexfn, name):
        return SqlIndex(self.dbpath, [], indexfn, name, self.stemmer_ids[name])

    def add_piece(self, piece, piece_path, explicit_repeats=False):
        with sqlite3.connect(self.dbpath) as conn:
            cursor = conn.cursor()
            cursor.execute("PRAGMA synchronous = OFF")
            cursor.execute("PRAGMA journal_mode = OFF")
            piece_id = None
            if explicit_repeats:
                try:
                    piece = piece.expandRepeats()
                except ExpanderException:
                    print("\tUnable to expand piece. Continuing with original")
            for part in get_part_details(piece):
                piece_name = part.piece
                part_name = part.name
                if not piece_id:
                    piece_id = self.ensure_piece(piece_path, piece_name, conn, cursor)
                part_id = self.ensure_part(piece_id, part_name, conn, cursor)
                snippets = list(get_snippets_for_part(part))
                snippet_ids = self.ensure_snippets(snippets, piece_id, part_id, conn, cursor)
                for idx in self.indexes.values():
                    idx.add_snippets(snippets, snippet_ids, conn, cursor)
            cursor.close()

    @staticmethod
    def ensure_db(conn):
        """
        Ensure base tables setup
            :param conn: Connection to sqlite instance
        """
        cursor = conn.cursor()
        cursor.execute("""CREATE TABLE IF NOT EXISTS pieces (id INTEGER PRIMARY KEY ASC,
                                                path TEXT NOT NULL,
                                                name TEXT NOT NULL
                        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS parts (id INTEGER PRIMARY KEY ASC,
                                                piece_id INTEGER NOT NULL,
                                                name TEXT NOT NULL,
                        FOREIGN KEY (piece_id) REFERENCES pieces(id))""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS stemmers (id INTEGER PRIMARY KEY ASC,
                                                    name TEXT NOT NULL
                        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS snippets (id INTEGER PRIMARY KEY ASC,
                                                    piece_id INTEGER NOT NULL,
                                                    part_id INTEGER NOT NULL,
                                                    offset INTEGER NOT NULL,
                                                    FOREIGN KEY (part_id) REFERENCES parts(id),
                                                    FOREIGN KEY (piece_id) REFERENCES pieces(id),
                                                    CONSTRAINT unique_snippet UNIQUE (piece_id, part_id, offset)
                        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS stems (id INTEGER PRIMARY KEY ASC,
                                                stemmer_id INTEGER NOT NULL,
                                                stem TEXT NOT NULL,
                                                FOREIGN KEY (stemmer_id) REFERENCES stemmers(id),
                                                CONSTRAINT unique_stem UNIQUE (stemmer_id, stem)
                        )""")
        cursor.execute("""CREATE TABLE IF NOT EXISTS entries (id INTEGER PRIMARY KEY ASC,
                                                stem_id INTEGER NOT NULL,
                                                snippet_id INTEGER NOT NULL,
                                                FOREIGN KEY (stem_id) REFERENCES stems(id),
                                                FOREIGN KEY (snippet_id) REFERENCES snippets(id),
                                                CONSTRAINT unique_entry UNIQUE (stem_id, snippet_id)
                        )""")
        conn.commit()
        cursor.execute("""CREATE INDEX IF NOT EXISTS piece_path_idx ON pieces(path)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS stemmer_name_idx ON stemmers(name)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS snippet_piece_idx ON snippets(piece_id)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS snippet_part_idx ON snippets(part_id)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS part_piece_idx ON parts(piece_id)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS stem_stemmer_idx ON stems(stemmer_id)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS stem_stem_idx ON stems(stem)""")
        cursor.execute("""CREATE INDEX IF NOT EXISTS entry_stem_idx ON entries(stem_id)""")

    @staticmethod
    def ensure_stemmers(stemmers, conn):
        """
        Ensure the given stemmers are included in the stemmers table
            :param stemmers: Stemmers to verify
            :param conn: Connection to sqlite instance
        """
        cursor = conn.cursor()
        stemmer_ids = {}
        for stemmer_name in stemmers.keys():
            cursor.execute("SELECT id FROM stemmers WHERE name=?", (stemmer_name, ))
            results = cursor.fetchall()
            if not results:
                cursor.execute("INSERT INTO stemmers (name) VALUES (?)", (stemmer_name, ))
            stemmer_ids[stemmer_name] = cursor.lastrowid
        conn.commit()
        return stemmer_ids

    @staticmethod
    def ensure_piece(piece_path, piece_name, conn, cursor):
        """
        Ensure the given piece is included
            :param piece_path: Path to the piece
            :param piece_name: Name of the piece
            :param conn: Connection to sqlite instance
            :param cursor: Cursor to use
        """
        cursor.execute("SELECT id FROM pieces WHERE path=? AND name=? LIMIT 1",
                       (piece_path, piece_name)
                      )
        results = cursor.fetchall()
        if results:
            return results[0][0]
        cursor.execute("INSERT INTO pieces (path, name) VALUES (?, ?)",
                       (piece_path, piece_name)
                      )
        conn.commit()
        return cursor.lastrowid

    @staticmethod
    def ensure_part(piece_id, part_name, conn, cursor):
        """
        Ensure given part is included
            :param piece_id: Id of source piece
            :param part_name: Name of the part
            :param conn: Connection to sqlite instance
            :param cursor: Cursor to use
        """
        cursor.execute("SELECT id FROM parts WHERE piece_id=? AND name=? LIMIT 1", (piece_id, part_name))
        results = cursor.fetchall()
        if results:
            return results[0][0]
        cursor.execute("INSERT INTO parts (piece_id, name) VALUES (?, ?)", (piece_id, part_name))
        conn.commit()
        return cursor.lastrowid

    @staticmethod
    def ensure_snippet(snippet, piece_id, part_id, conn, cursor):
        """
        Ensure the given snippet is included
            :param snippet: Snippet instance to store
            :param piece_id: Id of source piece
            :param part_id: Id of source part
            :param conn: Connection to sqlite instance
            :param cursor: Cursor to use
        """
        cursor.execute("SELECT id FROM snippets WHERE piece_id=? AND part_id=? AND offset=?", (piece_id, part_id, snippet.offset))
        results = cursor.fetchall()
        if results:
            return results[0][0]
        cursor.execute("INSERT INTO snippets (piece_id, part_id, offset) VALUES (?, ?, ?)", (piece_id, part_id, snippet.offset))
        conn.commit()
        return cursor.lastrowid

    @staticmethod
    def ensure_snippets(snippets, piece_id, part_id, conn, cursor):
        """
        Ensure several snippets from the same piece and part are included
            :param snippets: List of snippets to include
            :param piece_id: Id of source piece
            :param part_id: Id of source part
            :param conn: Connection to sqlite instance
            :param cursor: Cursor to use
        """
        values = [(piece_id, part_id, snippet.offset) for snippet in snippets]
        cursor.executemany("INSERT OR IGNORE INTO snippets (piece_id, part_id, offset) VALUES (?, ?, ?)", values)
        conn.commit()
        cursor.execute("SELECT id FROM snippets WHERE piece_id=? AND part_id=?", (piece_id, part_id))
        return [r[0] for r in cursor.fetchall()]

    def get_number_of_pieces(self):
        """
        Get the total number of pieces
            :param self:
        """
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        cursor.execute("""SELECT count(*) FROM pieces""")
        result = cursor.fetchone()
        conn.close()
        return result[0]

    def lookup(self, snippet):
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        return super().lookup(snippet, conn, cursor)

    def raw_query(self, query, *args):
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        return super().raw_query(query, conn, cursor, *args)

    def corpus_size(self):
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        cursor.execute("SELECT count(*) FROM pieces")
        result = cursor.fetchone()
        return result[0]

    def piece_by_id(self, piece_id):
        """
        Lookup a single piece by ID
            :param self:
            :param piece_id: Id of the piece to retrieve
        """
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM pieces WHERE pieces.id=?", (piece_id, ))
        return cursor.fetchall()

    def pieces(self):
        """
        Return basic information on all pieces
            :param self:
        """
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        cursor.execute("SELECT name, path, id FROM pieces")
        return cursor.fetchall()

    def stemmers(self):
        """
        Return basic information on all stemmers
            :param self:
        """
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM stemmers")
        return cursor.fetchall()

    def graders(self):
        """
        Return a sequence of the grading methods supported
            :param self:
        """
        return self.grader_methods.keys()

    def info(self):
        """
        Return general information about the data in FIRMS instance
            :param self:
        """
        tables = ["entries", "pieces", "parts", "snippets", "stems", "stemmers"]
        results = {}
        conn = sqlite3.connect(self.dbpath)
        cursor = conn.cursor()
        for table in tables:
            print("Querying table %s" % table)
            cursor.execute("SELECT count(*) FROM %s" % table)
            results[table] = cursor.fetchone()[0]
        return results

class SqlIndex(FirmIndex):
    def __init__(self, dbpath, snippets, keyfn, name, stemmer_id):
        self.dbpath = dbpath
        self.stemmer_id = stemmer_id
        super().__init__(snippets, keyfn, name)
    
    def ensure_stem(self, stemmer_id, stem, conn, cursor):
        cursor.execute("SELECT id FROM stems WHERE stemmer_id=? AND stem=? LIMIT 1", (stemmer_id, stem))
        results = cursor.fetchall()
        if results:
            return results[0][0]
        cursor.execute("INSERT INTO stems (stemmer_id, stem) VALUES (?, ?)", (stemmer_id, stem))
        conn.commit()
        return cursor.lastrowid

    def ensure_entry(self, stem_id, snippet_id, conn, cursor):
        cursor.execute("SELECT id FROM entries WHERE stem_id=? AND snippet_id=? LIMIT 1", (stem_id, snippet_id))
        results = cursor.fetchall()
        if results:
            return results[0][0]
        cursor.execute("INSERT INTO entries (stem_id, snippet_id) VALUES (?, ?)", (stem_id, snippet_id))
        conn.commit()
        return cursor.lastrowid
    
    def ensure_entries(self, stem_ids, snippet_ids, conn, cursor):
        values = list(zip(stem_ids, snippet_ids))
        cursor.executemany("INSERT OR IGNORE INTO entries (stem_id, snippet_id) VALUES (?, ?)", values)
        conn.commit()
        return [r[0] for value in values for r in cursor.execute("SELECT id FROM entries WHERE stem_id=? AND snippet_id=? LIMIT 1", value)]

    def ensure_stems(self, stemmer_id, stems, conn, cursor):
        values = [ (stemmer_id, stem) for stem in stems ]
        cursor.executemany("INSERT OR IGNORE INTO stems (stemmer_id, stem) VALUES (?, ?)", values)
        conn.commit()
        return [r[0] for value in values for r in cursor.execute("SELECT id FROM stems WHERE stemmer_id=? AND stem=? LIMIT 1", value)]

    def add_snippets(self, snippets, snippet_ids, conn, cursor):
        stems = [self.keyfn(snippet)[0] for snippet in snippets]
        stem_ids = self.ensure_stems(self.stemmer_id, stems, conn, cursor)
        return self.ensure_entries(stem_ids, snippet_ids, conn, cursor)

    def add_snippet(self, snippet, snippet_id, conn, cursor):
        stems = self.keyfn(snippet)
        for stem in stems:
            stem_id = self.ensure_stem(self.stemmer_id, stem, conn, cursor)
            entry_id = self.ensure_entry(stem_id, snippet_id, conn, cursor)
        return entry_id

    def lookup(self, snippet, conn, cursor):
        cursor.arraysize = 1000
        results = []
        stems = self.keyfn(snippet)
        for stem in stems:
            cursor.execute("""SELECT snippets.id, pieces.name, snippets.part_id as part, snippets.offset, stems.id, pieces.path, pieces.id FROM snippets
                            JOIN entries ON entries.snippet_id=snippets.id
                            JOIN stems ON stems.id=entries.stem_id
                            JOIN pieces ON pieces.id=snippets.piece_id
                            WHERE stems.stem=?
                            AND stems.stemmer_id=?""", (stem, self.stemmer_id))
            result = cursor.fetchmany()
            while result:
                results.append([ {'id': r[0], 'piece': r[5], 'part': r[2], 'offset': r[3], 'stem': r[4], 'path': r[5], 'piece_id': r[6]} for r in result ])
                result = cursor.fetchmany()

        return list(chain.from_iterable(results))



from setuptools import setup

setup(
    name='firms',
    version='0.0.6',
    description='Fuzzy Information Retrieval for Musical Scores',
    url='https://github.com/axiomabsolute/cs410-information-retrieval',
    download_url='https://github.com/axiomabsolute/cs410-information-retrieval/archive/0.0.6.tar.gz',
    author='Devon Terrell',
    author_email='devon@axioms.me',
    license='MIT',
    packages=['firms'],
    python_requires=">=3",
    entry_points={
        'console_scripts': ['firms=firms.command_line:main'],
    },
    install_requires=[
        'music21',
        'tabulate',
        'click',
        'scipy'
    ])



"""
Collection of models and functions for interacting with them.
"""

from collections import defaultdict, namedtuple
from abc import ABCMeta, abstractmethod
import os

import music21

# A part of a musical score, represented by a music21 stream
# and lineage information
Part = namedtuple('Part', ['piece', 'name', 'part'])

# A single query match row
LookupMatch = namedtuple('LookupMatch', [
    'id', 'piece', 'part', 'offset', 'stem', 'path', 'piece_id'
])

# A LookupMatch and the stemmer name that caused the match
GraderMatch = namedtuple('GraderMatch', ['stemmer', 'lookup_match'])

# A single result from grading a piece
GraderResult = namedtuple('GraderResult', ['piece', 'grade', 'meta'])

def flatten(toflatten):
    """
    Flattens nested iterable by one level
        :param toflatten: Sequence of nested sequences to flatten
    """
    return [item for sublist in toflatten for item in sublist]

def get_part_details(general_stream):
    """
    Gets a tuple of title, partName, and part for each part in a list of pieces
    """
    for piece in general_stream.recurse(classFilter=music21.stream.Score, skipSelf=False):
        piece_title = (piece and piece.metadata and piece.metadata.title) or "Untitled"
        for idx, part in enumerate(piece.recurse().parts):
            yield Part(piece_title, part.partName or "Part %s" % idx, part)

def get_notes_and_rests(part):
    """
    Given a part, extracts a flattened list of all notes in the part
    part - the part to extract
    """
    return list(part.recurse().notesAndRests)

def get_snippets_for_piece(piece_name, part_name, notes, snippet_length):
    """
    Generate a sequence of all snippets from a flat sequence of notes
        :param piece_name: Name of the source piece
        :param part_name: Name of the source part
        :param notes: Seq of notes
        :param snippet_length: Desired length of each snippet
    """
    return (Snippet(piece_name, part_name, notes[i: i+snippet_length], i) for i in range(0, 1 + len(notes) - snippet_length))

def get_snippets_for_part(part):
    """
    Generate all snippets for a part
    Can expand repeated sections by converting the part to MIDI and back. May be slow.
        :param part: The music21 part to generate snippets from
    """
    return get_snippets_for_piece(part.piece, part.name, get_notes_and_rests(part.part), 5)

class IRSystem(metaclass=ABCMeta):
    """
    A complete IR system that defines operations in terms of abstract FirmsIndex instances
        :param metaclass=ABCMeta: Abstract MetaClass
    """
    def __init__(self, index_methods, graders, piece_paths, rebuild=True):
        """
        Constructor
            :param self:
            :param index_methods: Dictionary of stemmers
            :param graders: Dictionary of graders
            :param piece_paths: List of file paths to pieces
            :param rebuild=True: Boolean flag erases existing FIRMS index if true
        """
        self.index_methods = index_methods
        self.grader_methods = graders
        self.indexes = {k:self.make_empty_index(v, k) for k, v in index_methods.items()}
        if rebuild:
            for idx, piece_path in enumerate(piece_paths):
                print("Adding piece #%s: %s" % (idx, piece_path))
                piece = music21.corpus.parse(piece_path)
                self.add_piece(piece, piece_path)

    @abstractmethod
    def add_piece(self, piece, piece_path):
        """
        Add a single piece to FIRMS
            :param self:
            :param piece: Music21 stream representing the piece
            :param piece_path: Original path to the piece
        """
        pass

    @abstractmethod
    def make_empty_index(self, indexfn, name):
        """
        Create a new, empty
            :param self:
            :param indexfn: Function defining how a single snippet is indexed
            :param name: Name of the indexing method
        """
        pass

    def lookup(self, snippet, *args):
        """
        Lookup a single snippet
            :param self:
            :param snippet: Music21 stream representing the snippet
            :param *args: Additional arguments, passed on to individual index lookups
        """
        snippets_by_index_type = {index_name: index.lookup(snippet, *args) for index_name, index in self.indexes.items()}
        return {scorer_name: scorer(snippets_by_index_type) for scorer_name, scorer in self.grader_methods.items()}

    @abstractmethod
    def corpus_size(self):
        """
        Returns the number of pieces in FIRMS instance
            :param self:
        """
        pass

    def raw_query(self, query, *args):
        """
        Perform a query without aggregating and grading results
            :param self:
            :param query: Query represented by a Music21 stream
            :param *args: Extra arguments passed on to index lookup methods
        """
        for grader in self.grader_methods.values():
            grader.zero()
        query_stream = None
        try:
            assert 'Stream' in query.classSet or 'StreamIterator' in query.classSet
            query_stream = query
        except AssertionError:
            query_stream = music21.tinyNotation.Converter.parse(query)
        query_part = Part("query", "query", query_stream)
        # This needs to be a list because it gets iterated over for every index type
        query_snippets = list(get_snippets_for_part(query_part))
        for index_name, index in self.indexes.items():
            for snippet in query_snippets:
                lookup_results = index.lookup(snippet, *args)
                for grader in self.grader_methods.values():
                    grader.aggregate([GraderMatch(stemmer=index_name, lookup_match=lookup_result) for lookup_result in lookup_results])

    def query(self, query, *args):
        """
        Perform a query, aggregate, and rank results
            :param self:
            :param query: Query represented by Music21 stream
            :param *args: Additional args passed on to raw_query, then to individual index queries
        """
        corpus_size = self.corpus_size()
        self.raw_query(query, *args)
        grades_by_grader = {grader_name: grader.grade(corpus_size) for grader_name, grader in self.grader_methods.items()}
        return grades_by_grader

class Snippet:
    """
    Represents a short snippet of a musical work as a list of note values, along with lineage
    information for piece and part.
    """
    def __init__(self, piece_name, part, notes, offset):
        """
        Constructor
            :param self:
            :param piece_name: Name of the originating piece
            :param part: Name of the originating part
            :param notes: List of music21 notes
            :param offset: Number of notes offset from the start of the part
        """
        self.piece = piece_name
        self.part = part
        self.notes = notes
        self.offset = offset
    
    def simple_line(self):
        """
        Returns a simplified string representation of the pitches of notes in a snippet.
        Used for debugging
            :param self:
        """
        return [
            note.pitch.nameWithOctave if note.isNote else
            '[%s]' % ' '.join([pitch.nameWithOctave for pitch in note.pitches]) if note.isChord else
            'rest%s' % note.duration.quarterLength
            for note in self.notes
        ]
        
    def __repr__(self):
        """
        String representation, based on simple_line and properties
            :param self:
        """
        return "Snippet(%s, %s, %s, %s)" % (self.piece, self.part, self.offset, self.simple_line())

class FirmIndex(metaclass=ABCMeta):
    """
    A single stemming method
        :param metaclass=ABCMeta: Abscract MetaClass
    """
    def __init__(self, snippets, keyfn, name=""):
        """
        Constructor
            :param self:
            :param snippets: List of snippets to add to index
            :param keyfn: Stemming method
            :param name="": Name of the stemming method
        """
        self.index = defaultdict(set)
        self.keyfn = keyfn
        self.name = name
        for snippet in snippets:
            self.add_snippet(snippet)

    @abstractmethod
    def add_snippet(self, snippet, *args):
        """
        Add a single snippet to the index
            :param self:
            :param snippet: Snippet to add
            :param *args: Arbitrary additional arguments
        """
        pass

    @abstractmethod
    def lookup(self, snippet, *args):
        """
        Look up a single snippet and return a list of LookupMatch tuples
            :param self:
            :param snippet: Snippet to lookup
            :param *args: Arbitrary extra args
        """
        pass

class Grader(metaclass=ABCMeta):
    """
    An implementation of a LookupMatch aggregation, grading, and ranking method.
    This is a *stateful* object, allowing results to be incrementally collected
    before aggregation.
        :param metaclass=ABCMeta: Abstract MetaClass
    """
    def __init__(self):
        self.zero()

    @abstractmethod
    def zero(self):
        """
        Reset the grader's aggregator
        """
        pass

    @abstractmethod
    def grade(self, number_of_pieces):
        """
        Compute grades for the pieces currently stored
            :param self:
            :param number_of_pieces: The total number of pieces in the corpus
        """
        pass

    @abstractmethod
    def aggregate(self, matches):
        """
        Add a set of results to the grader's aggregator
            :param self:
            :param matches: List of LookupMatch tuples to add to the aggregator
        """
        pass
